import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from .gui.gui_compare_motors import CompareMotors
from .gui.gui_single_motor_perf import SingleMotorPerfTab
from .gui.main_window import MainWindow

from .ressources import DEFAULT_LIBRARY
import pkg_resources

def nemo_main(is_unit_test = False):
    main = MainWindow()
    main.window.connect("delete-event", Gtk.main_quit)
    main.window.set_default_size(400, 300)

    c_tab = CompareMotors()
    main.add_tab(c_tab)
    main.add_tab(SingleMotorPerfTab())
    for tab in main.tabs:
        tab.update_library(DEFAULT_LIBRARY)
    c_tab.add_motor()

    version = pkg_resources.require("nemo_bldc")[0].version
    main.window.set_title(f"Nemo - {version}")
    main.window.show_all()
    if is_unit_test:
        for t in main.tabs:
            t.update_plot()
    else:
        Gtk.main()

if __name__ == "__main__":
    nemo_main()





import typing as tp
import numpy as np
from matplotlib.figure import Figure

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib, GdkPixbuf

from .widget_motor_creation import DisplayMotor
from .abstract_tab import AbstractTab
from .utils import *
from ..ressources import get_ressource_path


def motor_to_treeview(mot):
    return [color_to_pixbuf(mot.get_gtk_color()),
            mot.name,
            f"{1 / mot.K_m_art**2:.4f}",
            f"{mot.kt_q_art:.3f}",
            f"{mot.nominal_power:.1f}"]

class CompareMotors(AbstractTab):
    '''
    GUI tab: compare two motors
    '''
    def __init__(self):
        builder = Gtk.Builder()
        builder.add_from_file(get_ressource_path("compare_tab.glade"))
        super().__init__("Compare", builder.get_object("side_box"))

        # Get the widgets we need from the builder.
        self.compare_entry_R = builder.get_object("compare_R")
        self.compare_entry_L = builder.get_object("compare_L")
        self.compare_entry_ke = builder.get_object("compare_ke")
        self.compare_entry_I = builder.get_object("compare_I")
        self.compare_entry_n = builder.get_object("compare_n")

        self.mpl_fig = Figure(figsize=(5, 4), dpi=100)
        ax = self.mpl_fig.add_subplot()
        self.mpl_fig.subplots_adjust(left=0.2, bottom=0.15, right=0.97, top=0.97, wspace=0, hspace=0)
        def format_coord(x, y):
            return f"Velocity: {x:.1f}rad/s ({x * 30 / np.pi:.1f}rpm), Torque: {y:.1f}Nm"
        ax.format_coord = format_coord

        self.configure_plot(self.mpl_fig)

        self.tree_view = builder.get_object("tree_view")
        self.tree_view.get_selection().connect("changed", self.tree_changed)
        self.motor_compare_list = builder.get_object("motor_compare_list")

        self.selected_motor = None
        self.motors = []

        box = builder.get_object("side_box")
        box.pack_start(self.motor_widget.top_frame, False, False, 0)
        box.reorder_child(self.motor_widget.top_frame, 2)
        self.motor_widget.connect("motor_updated", self.motor_updated)
        self.motor_widget.connect("motor_name_updated", self.motor_name_updated)

        self.motorList = builder.get_object("motorList")
        self.n_motor_created = 0

        self.grid_ctes = builder.get_object("grid_ctes")
        self.grid_column = []
        self.grid_sep = [Gtk.Separator() for _ in DISPLAYED_CTES]

        grid_labels = builder.get_object("grid_labels")
        for i, (field, unit, _) in enumerate(DISPLAYED_CTES):
            grid_labels.insert_row(2 * i + 1)
            grid_labels.attach(Gtk.Label(label=field), 0, 2 * i + 1, 1 ,1)
            grid_labels.attach(Gtk.Label(label=unit), 1, 2 * i + 1, 1, 1)
            grid_labels.insert_row(2 * i + 2)
            grid_labels.attach(Gtk.Separator(), 0, 2 * i + 2, 2, 1)

        builder.connect_signals(self)

    def param_update(self):
        for i, m in enumerate(self.motors):
            self.grid_column[i][0].set_markup(f"<b>{m.name}</b>")
            for (_, _, func), label in zip(DISPLAYED_CTES, self.grid_column[i][1:]):
                label.set_text(func(m))
        self.plot_need_update()

    def update_plot(self):
        plot_caracteristics(self.mpl_fig.gca(), self.motors, [m.color for m in self.motors])
        self.mpl_fig.canvas.draw()

    def motor_updated(self, *args):
        if self.selected_motor is not None:
            self.motors[self.selected_motor] = self.motor_widget.motor
            self.motor_compare_list[self.selected_motor][2] = f"{1 / self.motor_widget.motor.K_m_art**2:.3f}"
            self.motor_compare_list[self.selected_motor][3] = f"{self.motor_widget.motor.kt_q_art:.3f}"
            self.motor_compare_list[self.selected_motor][4] = f"{self.motor_widget.motor.nominal_power:.1f}"
            self.param_update(),

    def tree_changed(self, selection):
        # Get selection id
        model, treeiter = selection.get_selected()
        if treeiter is None:
            self.selected_motor = None
        else:
            self.selected_motor = int(str(model[treeiter].path))
            self.motor_widget.set_motor(self.motors[self.selected_motor])

    def add_motor(self, *args):
        name = list(self.motor_widget.motor_library.keys())[0]
        m = self.motor_widget.motor_library[name]
        mot = DisplayMotor(m, name,f"C{self.n_motor_created}")
        self.n_motor_created += 1
        self.motors.append(mot)
        self.motor_compare_list.append(motor_to_treeview(mot))

        # Create new column.
        nc = len(self.grid_column)
        self.grid_ctes.insert_column(2 * nc)
        self.grid_ctes.attach(Gtk.Separator(), 2 * nc, 0, 1, 2 * len(DISPLAYED_CTES) + 1)
        self.grid_ctes.insert_column(2 * nc + 1)
        self.grid_column.append([Gtk.Label(label='label')] + [Gtk.Label(label='label') for _ in DISPLAYED_CTES])
        for i, label in enumerate(self.grid_column[-1]):
            self.grid_ctes.attach(label, 2 * nc + 1, 2 * i - 1, 1, 1)
            self.grid_ctes.attach(Gtk.Separator(), 2 * nc + 1, 2 * i, 1, 1)

        self.grid_ctes.show_all()

        # Set cursor to new value
        self.tree_view.set_cursor(len(self.motors) - 1)

    def remove_motor(self, *args):
        if self.selected_motor is not None:
            idx = self.selected_motor
            del self.motors[idx]
            del self.grid_column[idx]
            self.motor_compare_list.clear()
            for mot in self.motors:
                self.motor_compare_list.append(motor_to_treeview(mot))
            self.tree_view.set_cursor(Gtk.TreePath())
            self.grid_ctes.remove_column(2 * idx)
            self.grid_ctes.remove_column(2 * idx)

    def motor_name_updated(self, *args):
        if self.selected_motor is not None:
            new_name = self.motor_widget.motor.name
            self.motors[self.selected_motor].name = new_name
            self.motor_compare_list[self.selected_motor][1] = new_name
            self.grid_column[self.selected_motor][0].set_text(new_name)



import matplotlib.colors as mc
import colorsys
import numpy as np
import functools
import json

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk, GdkPixbuf

from ..physics.motor import Motor

DISPLAYED_CTES = [('Km, articular', 'Nm/\u221AW', lambda m: f"{m.K_m_art:.3f}"),
                  ('', 'W /Nm\u00B2', lambda m: f"{1 / m.K_m_art**2:.3f}"),
                  ('Power (non defluxing)', 'W', lambda m: f"{m.w_max_at_max_torque * m.iq_max * m.kt_q_art:.1f}"),
                  ('Ke phase-phase', 'V/rad.s', lambda m: f"{m.ke_phasetophase:.3f}"),
                  ('', 'V/krpm', lambda m: f"{m.ke_phasetophase * 1000 * np.pi / 30:.3f}"),
                  ('KV', 'rpm/V', lambda m: f"{1 / (m.ke_phasetophase * np.pi / 30):.0f}"),
                  ('Ktq, articular', 'Nm/A', lambda m: f"{m.kt_q_art:.3f}"),
                  ('Max RMS current', 'A', lambda m: f"{m.i_rms_max:.1f}"),
                  ('Max torque, articular', 'Nm', lambda m: f"{m.tau_max:.1f}"),
                  ('No load speed, articular', 'rad/s', lambda m: f"{m.w_max_no_load:.1f}"),
                  ('', 'rpm', lambda m: f"{30 / np.pi * m.w_max_no_load:.0f}"),
                  ('Max speed @ max torque', 'rad/s', lambda m: f"{m.w_max_at_max_torque:.1f}"),
                  ('', 'rpm', lambda m: f"{30 / np.pi * m.w_max_at_max_torque:.0f}")]

def gdk_rgba_to_tuple(color: Gdk.RGBA):
    return (color.red, color.green, color.blue, color.alpha)

def mpl_to_gdk_rgba(color):
    return Gdk.RGBA(*mc.to_rgba(color))

def color_to_pixbuf(color: Gdk.RGBA):
    pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, False, 8, 50, 25)
    fillr = int(color.red * 255) << 24
    fillg = int(color.green * 255) << 16
    fillb = int(color.blue * 255) << 8
    fillcolor = fillr | fillg | fillb | 255
    pixbuf.fill(fillcolor)
    return pixbuf

def lighten_color(color, amount=0.5):
    """
    Lightens the given color by multiplying (1-luminosity) by the given amount.
    Input can be matplotlib color string, hex string, or RGB tuple.

    Examples:
    >> lighten_color('g', 0.3)
    >> lighten_color('#F034A3', 0.6)
    >> lighten_color((.3,.55,.1), 0.5)
    """
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])

def plot_motor_caracteristic(ax: "matplotlib.axis", motor, color, four_quadrants=False, linewidth=2, linestyle='-'):
    tau = np.arange(0, motor.tau_max, 0.01)
    w_no_deflux = np.array([motor.compute_max_speed_no_deflux(t) for t in tau] + [motor.w_max_at_max_torque, 0])
    ax.plot(w_no_deflux, list(tau) + [motor.tau_max, motor.tau_max], color=color, linewidth=linewidth, linestyle=linestyle)
    # Defluxing
    w_deflux = np.array([motor.compute_max_speed_deflux(t) for t in tau] + [motor.w_max_at_max_torque])
    ax.plot(w_deflux, list(tau) + [motor.tau_max], color=lighten_color(color), linewidth=linewidth, linestyle=linestyle)
    # 4 quandrants.
    if four_quadrants:
        ax.plot(-w_deflux, tau, color=lighten_color(color), linewidth=linewidth, linestyle=linestyle)
        ax.plot(w_deflux, -tau, color=lighten_color(color), linewidth=linewidth, linestyle=linestyle)
        ax.plot(-w_deflux, -tau, color=lighten_color(color), linewidth=linewidth, linestyle=linestyle)

        ax.plot(w_no_deflux, list(-tau) + [-motor.tau_max, -motor.tau_max], color=color, linewidth=linewidth, linestyle=linestyle)
        ax.plot(-w_no_deflux, list(tau) + [motor.tau_max, motor.tau_max], color=color, linewidth=linewidth, linestyle=linestyle)
        ax.plot(-w_no_deflux, list(-tau) + [-motor.tau_max, -motor.tau_max], color=color, linewidth=linewidth, linestyle=linestyle)


def plot_caracteristics(ax,
                        motors,
                        colors,
                        margin=0,
                        four_quadrants=False,
                        plot_nominal=True,
                        secondary_y_axes=True):
        ax.clear()

        for m, c in zip(motors, colors):
            plot_motor_caracteristic(ax, m, c, four_quadrants)
            if plot_nominal:
                m_nom = Motor(1, 1, 1, 1, 1, 1, 48, 1)
                m_nom.copy(m)
                m_nom.update_constants(iq_max = m_nom.iq_nominal)
                plot_motor_caracteristic(ax, m_nom, c, four_quadrants, 2, 'dotted')

        # Adjust range
        if len(motors) > 0:
            i_m = max([m.tau_max for m in motors])
            w_m = max([m.w_max_no_load for m in motors])
            if four_quadrants:
                ax.set_xlim([-2 * w_m, 2 * w_m])
            else:
                ax.set_xlim([-margin * w_m, 2 * w_m])
            ax.set_xlabel("Articular speed (top rad/s, bottom rpm)")
            ax.secondary_xaxis(-0.07, functions=(lambda x: x * 30 / np.pi, lambda x: x * np.pi / 30))

            if four_quadrants:
                ax.set_ylim([-1.1 * i_m, 1.1 * i_m])
            else:
                ax.set_ylim([-margin * i_m, 1.1 * i_m])
            ax.set_ylabel("Torque (Nm)")
            # Create axes
            if secondary_y_axes:
                if len(motors) == 1:
                    ax.text(-0.1, 0.4, 'Quadrature current (A)', transform=ax.transAxes, rotation=90)
                else:
                    ax.text(-0.15, -0.03, 'Quadrature current (A)', transform=ax.transAxes)
                for i, m in enumerate(motors):
                    def direct(x, m):
                        return x / m.kt_q_art
                    def inv(x, m):
                        return x * m.kt_q_art
                    sa = ax.secondary_yaxis(-0.06 -0.04 * i , functions=(functools.partial(direct, m=m), functools.partial(inv, m=m)), color=colors[i])
            ax.grid(True)




# Modeling of a battery as voltage source + resistor
import typing as tp
import numpy as np

def get_battery_state(U_bat, R_bat, P):
    '''
    Return battery current and voltage, given a power input.
    The battery is modeled as a constant voltage source U_bat (typ. 48V), with
    a resistor R_bat in serie.
     - U_bat: battery voltage
     - R_bat: battery resistor
     - P: power drawn
    Return: U, I
    '''
    I = (U_bat - np.sqrt(U_bat**2 - 4 * R_bat * P)) / 2 / R_bat
    return U_bat - R_bat * I, I



import typing as tp
import numpy as np
import copy

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject, Gdk

import matplotlib.colors as mcolors
from ..physics.motor import Motor

from ..ressources import get_ressource_path

from .motor_creation_helper import MotorCreationHelper

class DisplayMotor(Motor):
    @staticmethod
    def FromJson(data):
        motor = Motor.FromDict(data)
        return DisplayMotor(motor, data["name"], data["color"])

    def __init__(self, motor, name, color):
        super().__init__(1, 1, 1, 1, 1, 1, 48, 1)
        super().copy(motor)

        self.name = name
        self.color = color

    def get_gtk_color(self):
        return Gdk.RGBA(*mcolors.to_rgba(self.color))

    def to_json(self):
        d = self.to_dict()
        d["name"] = self.name
        d["color"] = self.color
        return d

class MotorCreationWidget(GObject.Object):
    '''
    A widget for defining a motor from input parameters
    '''
    @GObject.Signal
    def motor_updated(self):
        pass

    @GObject.Signal
    def motor_name_updated(self):
        pass

    def __init__(self):
        GObject.GObject.__init__(self)
        builder = Gtk.Builder()
        builder.add_from_file(get_ressource_path("motor_creation_widget.glade"))
        self.top_frame = builder.get_object("top_frame")

        self.label_header = builder.get_object("header")

        self.entry_name = builder.get_object("entry_name")
        self.spin_R = builder.get_object("spin_R")
        self.spin_L = builder.get_object("spin_L")
        self.spin_ke = builder.get_object("spin_ke")
        self.spin_I = builder.get_object("spin_I")
        self.spin_np = builder.get_object("spin_np")
        self.spin_U = builder.get_object("spin_U")
        self.spin_rho = builder.get_object("spin_rho")
        self.spin_iqnom = builder.get_object("spin_iqnom")

        self.motor_box = builder.get_object("motor_box")
        self.motor_list = builder.get_object("motor_list")
        self.set_motor(DisplayMotor(Motor(1, 0.01, 0.0001, 1.0, 1.0, 1.0, 1.0, 1.0),
                                  "Name",
                                  "C0"))

        self.parent_callback = None
        builder.connect_signals(self)

    def set_motor(self, motor: DisplayMotor):
        self.motor = copy.copy(motor)
        self.entry_name.set_text(motor.name)
        self.spin_R.set_value(motor.R)
        self.spin_L.set_value(1000. * motor.L)
        self.spin_ke.set_value(motor.ke)
        self.spin_I.set_value(motor.iq_max)
        self.spin_iqnom.set_value(motor.iq_nominal)
        self.spin_iqnom.set_value(motor.iq_nominal)
        self.spin_np.set_value(2 * motor.np)
        self.spin_U.set_value(motor.U)
        self.spin_rho.set_value(motor.rho)
        self.motor_box.set_active(-1)
        self.input_updated()

    def input_updated(self, *args, **kargs):
        self.motor.update_constants(
            n = int(self.spin_np.get_value()),
            R = self.spin_R.get_value(),
            L = self.spin_L.get_value() / 1000.0,
            ke = self.spin_ke.get_value(),
            iq_max = self.spin_I.get_value(),
            iq_nominal = self.spin_iqnom.get_value(),
            U = self.spin_U.get_value(),
            reduction_ratio = self.spin_rho.get_value())
        self.emit("motor_updated")

    def name_updated(self, *args):
        self.motor.name = self.entry_name.get_text()
        self.emit("motor_name_updated")

    def box_updated(self, combo: tp.Any):
        tree_iter = combo.get_active_iter()
        if tree_iter is None:
            return
        name = combo.get_model()[tree_iter][0]

        self.motor.copy(self.motor_library[name])
        self.motor.name = name

        self.set_motor(self.motor)
        self.input_updated()

    def update_library(self, library):
        '''
        Update motor library.
        '''
        self.motor_library = library
        self.motor_list.clear()
        for n in self.motor_library:
            self.motor_list.append([n])
        self.motor_box.set_active(0)

    def ask_for_helper(self, button):
        helper = MotorCreationHelper(self.motor)
        result = helper.run()
        if result is not None:
            self.set_motor(DisplayMotor(result, self.motor.name, self.motor.color))




import typing as tp
from pathlib import Path
import numpy as np
import subprocess
import sys
import os

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk

from ..ressources import get_ressource_path, load_motor_library
from ..doc import get_doc_path

class MainWindow:
    '''
    GUI tab: compare two motors
    '''
    def __init__(self):
        builder = Gtk.Builder()
        builder.add_from_file(get_ressource_path("main_window.glade"))
        self.window = builder.get_object("window")
        self.window = builder.get_object("window")
        # self.window.
        self.notebook = builder.get_object("notebook")
        self.tabs = []
        self.switch_external_window = builder.get_object("switch_external_window")
        self.external_window = Gtk.Window()
        self.external_window.set_title("Nemo - External viewer")
        self.external_window.set_deletable(False)
        self.external_window.set_size_request(600, 800)
        self.external_window.connect("window-state-event", self.minimize_external_window)
        self.current_tab = None
        self.is_using_external_window = False
        builder.connect_signals(self)

    def add_tab(self, tab: "AbstractTab"):
        self.tabs.append(tab)
        self.notebook.append_page(tab.tab_widget, Gtk.Label(label=tab.name))

    def library_updated(self, fc_button):
        new_lib = load_motor_library(Path(fc_button.get_filename()))
        for tab in self.tabs:
            tab.update_library(new_lib)

    def _pdf_viewer_not_available(self, path):
        dialog = Gtk.Dialog(parent=self.window)
        dialog.add_button("Ok", 0)
        lab = Gtk.Label()
        lab.set_markup("<big>Oh no - PDF files cannot be opened in docker :( You can find this doc at:</big>")
        dialog.vbox.add(lab)

        ent = Gtk.Entry()
        ent.set_text(path)
        ent.set_editable(False)
        dialog.vbox.add(ent)
        dialog.vbox.set_spacing(10)

        dialog.set_size_request(800, -1)
        dialog.show_all()
        dialog.run()
        dialog.destroy()

    def show_math_doc(self, *args):
        if sys.platform == 'linux':
            try:
                subprocess.call(["xdg-open", get_doc_path("BrushlessMotorPhysics.pdf")])
            except FileNotFoundError:
                self._pdf_viewer_not_available(get_doc_path("BrushlessMotorPhysics.pdf"))
        else:
            os.startfile(get_doc_path("BrushlessMotorPhysics.pdf"))

    def show_user_manual(self, *args):
        if sys.platform == 'linux':
            try:
                subprocess.call(["xdg-open", get_doc_path("user_manual.pdf")])
            except FileNotFoundError:
                self._pdf_viewer_not_available(get_doc_path("user_manual.pdf"))
        else:
            os.startfile(get_doc_path("user_manual.pdf"))

    def switch_plot_window(self, widget, state):
        self.is_using_external_window = state
        if self.is_using_external_window:
            self.move_current_tab_plot_to_window()
            self.external_window.show_all()
        else:
            self.return_plot_to_current_tab()
            self.external_window.hide()

    def move_current_tab_plot_to_window(self):
        self.current_tab.tab_widget.remove(self.current_tab.scroll_widget)
        self.external_window.add(self.current_tab.scroll_widget)

    def return_plot_to_current_tab(self):
        self.external_window.remove(self.current_tab.scroll_widget)
        self.current_tab.tab_widget.pack2(self.current_tab.scroll_widget, True, False)

    def change_tab(self, notebook, page, page_num):
        # If using external window, rearrange widgets
        if self.is_using_external_window:
            self.return_plot_to_current_tab()
            self.current_tab = self.tabs[page_num]
            self.move_current_tab_plot_to_window()
        self.current_tab = self.tabs[page_num]

    def minimize_external_window(self, widget, event):
        if self.is_using_external_window:
            if event.new_window_state & Gdk.WindowState.ICONIFIED > 0:
                self.switch_external_window.set_state(False)
                self.external_window.deiconify()



import typing as tp
import numpy as np
import copy

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject

from ..physics.motor import Motor
from ..ressources import get_ressource_path
from .utils import DISPLAYED_CTES

class MotorCreationHelper(GObject.Object):
    '''
    A widget for defining a motor from input parameters
    '''
    @GObject.Signal
    def motor_updated(self):
        pass

    @GObject.Signal
    def motor_name_updated(self):
        pass

    def __init__(self, motor):
        GObject.GObject.__init__(self)
        builder = Gtk.Builder()
        builder.add_from_file(get_ressource_path("motor_creation_helper.glade"))
        self.dialog = builder.get_object("dialog")
        self.dialog.set_title("Nemo - Motor creation helper")

        self.motor = Motor(1, 0.01, 0.0001, 1.0, 1.0, 1.0, 1.0, 1.0)

        self.toggle_winding = builder.get_object("toggle_winding")
        self.toggle_RL = builder.get_object("toggle_RL")
        self.toggle_current = builder.get_object("toggle_current")
        self.toggle_poles = builder.get_object("toggle_poles")

        self.box_mag = builder.get_object("box_mag")
        self.toggle_ke = builder.get_object("toggle_ke")
        self.toggle_kv = builder.get_object("toggle_kv")
        self.box_ke = builder.get_object("box_ke")
        self.box_kv = builder.get_object("box_kv")
        self.box_kt = builder.get_object("box_kt")

        self.toggle_ke_sp = builder.get_object("toggle_ke_sp")
        self.toggle_ke_pp = builder.get_object("toggle_ke_pp")
        self.toggle_ke_sp_rpm = builder.get_object("toggle_ke_sp_rpm")
        self.toggle_ke_pp_rpm = builder.get_object("toggle_ke_pp_rpm")
        self.toggle_kt_A = builder.get_object("toggle_kt_A")

        self.grid_labels = builder.get_object("grid_labels")
        self.ctes_labels = []
        for i, (field, unit, _) in enumerate(DISPLAYED_CTES):
            self.grid_labels.insert_row(2 * i + 1)
            self.grid_labels.attach(Gtk.Label(label=field), 0, 2 * i + 1, 1 ,1)
            self.grid_labels.attach(Gtk.Label(label=unit), 1, 2 * i + 1, 1, 1)
            self.grid_labels.insert_row(2 * i + 2)
            self.grid_labels.attach(Gtk.Separator(), 0, 2 * i + 2, 3, 1)
            l = Gtk.Label("")
            self.grid_labels.attach(l, 2, 2 * i + 1, 1, 1)
            self.ctes_labels.append(l)

        self.spin_R = builder.get_object("spin_R")
        self.spin_L = builder.get_object("spin_L")
        self.spin_ke = builder.get_object("spin_ke")
        self.spin_Inom = builder.get_object("spin_Inom")
        self.spin_Imax = builder.get_object("spin_Imax")
        self.spin_np = builder.get_object("spin_np")
        self.spin_U = builder.get_object("spin_U")

        self.spin_R.set_value(motor.R)
        self.spin_L.set_value(1000. * motor.L)
        self.spin_ke.set_value(motor.ke)
        self.spin_Inom.set_value(motor.iq_nominal)
        self.spin_Imax.set_value(motor.iq_max)
        self.spin_np.set_value(motor.np)
        self.spin_U.set_value(motor.U)

        self.label_R = builder.get_object("label_R")
        self.label_L = builder.get_object("label_L")
        self.label_ke = builder.get_object("label_ke")
        self.label_np = builder.get_object("label_np")
        self.label_inom = builder.get_object("label_inom")
        self.label_imax = builder.get_object("label_imax")

        builder.connect_signals(self)
        self.input_updated()

    def input_updated(self, *args, **kargs):
        '''
        Miscellaneous user input change: update motor and refresh display.
        '''
        is_delta = not self.toggle_winding.get_active()

        R = self.spin_R.get_value()
        L = self.spin_L.get_value() / 1000.0

        # Convert to single phase, star
        is_phase_to_phase = not self.toggle_RL.get_active()
        coeff = 0.5 if is_phase_to_phase else (1/3 if is_delta else 1)
        R *= coeff
        L *= coeff

        # Magnetic parameter
        val = self.spin_ke.get_value()
        if self.toggle_ke.get_active():
            if self.toggle_ke_sp.get_active():
                ke = val * 1/np.sqrt(3) if is_delta else 1
            elif self.toggle_ke_pp.get_active():
                ke = val / np.sqrt(3)
            elif self.toggle_ke_sp_rpm.get_active():
                ke = val * 30 / np.pi * 1/np.sqrt(3) if is_delta else 1
            else:
                ke = val * 30 / np.pi / np.sqrt(3)
        elif self.toggle_kv.get_active():
            ke = 1 / val * 30 / np.pi / np.sqrt(3)
        else:
            ktq = val * (1 if self.toggle_kt_A.get_active() else 1 / np.sqrt(2))
            ke = 2.0 / 3.0 * ktq

        npoles = int(self.spin_np.get_value())
        if not self.toggle_poles.get_active():
            npoles *= 2

        current_coeff = 1 if self.toggle_current.get_active() else np.sqrt(2)

        self.motor.update_constants(
            n = npoles,
            R = R,
            L = L,
            ke = ke,
            iq_max = current_coeff * self.spin_Imax.get_value(),
            iq_nominal = current_coeff * self.spin_Inom.get_value(),
            U = self.spin_U.get_value())

        # Update GUI
        self.label_R.set_text(f"{self.motor.R:.4f}")
        self.label_L.set_text(f"{1000 * self.motor.L:.4f}")
        self.label_ke.set_text(f"{self.motor.ke:.4f}")
        self.label_np.set_text(f"{int(2 * self.motor.np)}")
        self.label_inom.set_text(f"{self.motor.iq_nominal:.2f}")
        self.label_imax.set_text(f"{self.motor.iq_max:.2f}")

        for i, (_, _, f) in zip(self.ctes_labels, DISPLAYED_CTES):
            i.set_text(f(self.motor))

    def mag_updated(self, *args):
        '''
        User updated the magnetic parameter choice (Ke, Kv, Kt)
        '''
        if self.toggle_ke.get_active():
            act = self.box_ke
        elif self.toggle_kv.get_active():
            act = self.box_kv
        else:
            act = self.box_kt
        self.box_mag.remove(self.box_mag.get_children()[-1])
        self.box_mag.pack_end(act, True, True, 0)
        self.input_updated()

    def run(self):
        """
        Run the motor creation helper ; return the new motor, or None
        if the user canceled.
        """
        self.dialog.show_all()
        result = self.dialog.run()
        self.dialog.hide()
        if result == int(Gtk.ResponseType.OK):
            return self.motor
        return None


import pkg_resources
import json

from ..physics.motor import Motor

def get_ressource_path(filename:str):
    return pkg_resources.resource_filename(__name__, filename)

def load_motor_library(source_file: str):
    '''
    Load a motor library json file.
    '''
    library = {}
    with open(source_file, "r") as f:
        data = json.load(f)
        for k, d in data.items():
            try:
                library[k] = Motor.FromDict(d)
            except KeyError:
                print(f"Warning: failed to load {k}")
    return library

DEFAULT_LIBRARY = load_motor_library(get_ressource_path("motor_library.json"))


import typing as tp
import numpy as np

class Motor:
    '''
    A class to represent a PMSM motor, implementing all relevant computations
    '''
    @staticmethod
    def FromDict(data):
        m = Motor(1, 1, 1, 1, 1, 1, 48, 1)
        m.update_constants(n = data['np'],
                           R = data['R'],
                           L = data['L'] / 1000.0,
                           ke = data['ke'],
                           iq_max = data['i_quadrature_max'],
                           iq_nominal = data.get('i_quadrature_nominal', data['i_quadrature_max']),
                           U = data['U'],
                           reduction_ratio = data['reduction_ratio'])
        return m

    def __init__(self, n:int, R:float, L:float, ke:float, iq_max:float, iq_nominal:float, U: float, reduction_ratio:float):
        '''
        Build a PMSM motor from the fundamental parameters.
            - n: number of poles
            - R: resistor, per phase, in Ohm
            - L: phase inductance, in H
            - ke: rotor flux in a phase
            - iq_max: maximum quadrature current (assuming no defluxing) in the motor
            - U: driver voltage
            - reduction_ratio: reduction ratio
        Note:
            - magnetic saturations not modelled
            - salliancy ratio is set to 1
        '''
        self.update_constants(n, R, L, ke, iq_max, iq_nominal, U, reduction_ratio)

    def to_dict(self):
        '''
        Store motore in dictionnary to save to json file.
        '''
        return {"np": 2 * self.np,
                "R": self.R,
                "L": self.L,
                "L": 1000.0 * self.L,
                "ke": self.ke,
                "i_quadrature_max": self.iq_max,
                "i_quadrature_nominal": self.iq_nominal,
                "U": self.U,
                "reduction_ratio": self.rho
                }

    def copy(self, other_motor:"Motor"):
        '''
        Copy the constants of other_motor onto self.
        '''
        self.update_constants(n = 2 * other_motor.np,
                              R = other_motor.R,
                              L = other_motor.L,
                              ke = other_motor.ke,
                              iq_max = other_motor.iq_max,
                              iq_nominal = other_motor.iq_nominal,
                              U = other_motor.U,
                              reduction_ratio = other_motor.rho)

    def update_constants(self,
                         n:int = None,
                         R:float = None,
                         L:float = None,
                         ke:float = None,
                         iq_max:float = None,
                         iq_nominal:float = None,
                         U: float = None,
                         reduction_ratio: float = None):
        '''
        Update the motor's constants (none to keep previous value).
        Note that updating the parameters should be done through this function
        only, otherwise the derived constants will be wrong !
        The parameters are the same as in the constructor.
        '''
        if n is not None:
            self.np = n / 2
        if R is not None:
            self.R = R
        if L is not None:
            self.L = L
        if ke is not None:
            self.ke = ke
        if iq_max is not None:
            self.iq_max = iq_max
        if iq_nominal is not None:
            self.iq_nominal = iq_nominal
        if U is not None:
            self.U = U
        if reduction_ratio is not None:
            self.rho = reduction_ratio

        self._compute_derived_constants()

    def _compute_derived_constants(self):
        '''
        Compute useful constants, derived from the fundamental parameters
        '''
        # See BrushlessMotorPhysics.pdf for more informations on these constants.
        # Torque-related constants
        self.kt_q_art = 3.0 / 2.0 * self.rho * self.ke
        self.i_rms_max = self.iq_max / np.sqrt(2)
        self.tau_max = self.kt_q_art * self.iq_max

        # Velocity-related constants
        self.ke_phasetophase = np.sqrt(3) * self.ke
        self.w_max_no_load = self.U / self.ke_phasetophase / self.rho
        self.w_max_at_max_torque = self.compute_max_speed_no_deflux(self.tau_max)

        # Power-related constants
        self.K_m_art = np.sqrt(2.0 / 3.0) * self.kt_q_art / np.sqrt(self.R)

        # Miscellaneous.
        self.r_deflux = self.np * self.L * self.iq_max / self.ke
        tau_n = self.kt_q_art * self.iq_nominal
        self.nominal_power = self.compute_max_speed_no_deflux(tau_n) * tau_n


    def __str__(self):
        return f"R: {self.R}Ohm, L: {self.L * 1000.0}mH, Phi: {self.ke}Wb, Iq_max: {self.iq_max}A, Np {self.np}, U {self.U}, reduction {self.rho}"

    def compute_max_speed_no_deflux(self, tau):
        '''
        Return the maximum articular speed, given articular torque, when not defluxing.
         - tau: input articular torque, Nm
        '''
        tau = np.asarray(tau)

        iq = tau / self.kt_q_art

        a = self.rho**2 * ((self.np * self.L * iq)**2 + self.ke**2)
        b = 2 * self.rho * self.R * self.ke * iq
        c = (self.R * iq)**2 - self.U**2 / 3

        return (-b + np.sqrt(b**2 - 4 * a * c)) / 2 / a

    def compute_defluxing_current(self, tau, w):
        '''
        Get defluxing current, in A, given articular torque and velocity.
        '''
        tau = np.asarray(tau)
        w = np.asarray(w)

        i_q = tau / self.kt_q_art

        # Compute the minimum amount of current that is needed to deflux:
        # this corresponds to the amount of current such that U = U_eff.
        a = self.R**2 + (self.rho * w * self.np * self.L)**2
        b = 2 * self.np * self.L * self.ke * (self.rho * w)**2
        c = (self.rho * w * self.np * self.L * i_q)**2 \
            + 2 * self.R * i_q * self.ke * self.rho * w \
            + self.R**2 * i_q**2 + (self.ke * self.rho * w)**2 \
            - self.U**2  / 3

        return np.minimum(0.0, (-b + np.sqrt(b**2 - 4 * a * c)) / 2 / a)

    def compute_max_speed_deflux(self, tau):
        '''
        Returns the maximum articular speed, given articular torque, with defluxing.
         - tau: input articular torque, Nm
        '''
        tau = np.asarray(tau)

        i_q = tau / self.kt_q_art
        i_d = np.maximum(-np.sqrt(np.maximum(0, 2 * self.i_rms_max**2 - i_q**2)),
                         - self.ke / self.np / self.L)

        a = (self.rho * self.np * self.L * i_q)**2  + self.rho**2 * (self.np * self.L * i_d + self.ke)**2
        b = 2 * self.rho * self.R * i_q * self.ke
        c = self.R**2 * (i_d**2 + i_q**2) - self.U**2  / 3
        return np.maximum((-b + np.sqrt(b**2 - 4 * a * c)) / 2 / a, self.compute_max_speed_no_deflux(tau))

    def compute_thermal_power(self, tau, w, force_no_defluxing=False):
        '''
            Compute the thermal power to reach a specific working point.
            The motor will deflux if needed, unless specified.
            Note: this function does not check that the point is feasible for
            the motor: if you ask for infinite torque, you get infinite power !
        '''
        tau = np.asarray(tau)
        w = np.asarray(w)

        i_q = tau / self.kt_q_art
        if force_no_defluxing:
            i_d = np.zeros(w.shape)
        else:
            i_d = self.compute_defluxing_current(tau, w)
        power = 3 / 2 * self.R * (i_d**2 + i_q**2)
        return power

    def get_power(self, w, tau):
        '''
        Return the total power (mecanical + thermal) required by the motor,
        (assuming no defluxing).
        '''
        return w * tau + 1 / self.K_m_art**2 * tau**2


from setuptools import setup, find_namespace_packages
from pathlib import Path

setup(
    name='nemo_bldc',
    version='1.0.0',
    description='A tool to evaluate and compare brushless motors',
    long_description=Path('README.md').read_text('utf8'),
    author="Matthieu Vigne",
    license="MIT",
    packages=find_namespace_packages('src'),
    package_dir={'': 'src'},
    package_data={
        "nemo_bldc.ressources": ["*.glade", "*.json"],
        "nemo_bldc.doc": ["*.pdf"]
    },
    install_requires=[
        "pycairo", # Can cause issues if not installed before PyGObject - use pip and not setuptools to run install
        "pytest",
        "scipy",
        "numpy",
        "matplotlib",
        "PyGObject"],
    entry_points={
          'console_scripts': [
             'nemo_bldc = nemo_bldc.nemo:nemo_main'
          ]
      },
    include_package_data=True,
    zip_safe=False)



import typing as tp
import numpy as np

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib, GdkPixbuf

from matplotlib.backends.backend_gtk3agg import (
    FigureCanvasGTK3Agg as FigureCanvas)

from matplotlib.backends.backend_gtk3 import (
    NavigationToolbar2GTK3 as NavigationToolbar)

from matplotlib.figure import Figure

from .widget_motor_creation import MotorCreationWidget
from ..ressources import get_ressource_path

class AbstractTab:
    '''
    A tab of the analysis tool
    '''
    def __init__(self, name, sidebar_widget):
        sidebar_widget.set_size_request(300, -1)
        self.name = name

        builder = Gtk.Builder()
        builder.add_from_file(get_ressource_path("abstract_tab.glade"))
        self.tab_widget = builder.get_object("paned")
        # When designing in glade, the scroll widget is always the first
        # child: move it in second position.
        child = self.tab_widget.get_child1()
        self.tab_widget.remove(child)
        self.tab_widget.pack1(sidebar_widget, False, False)
        self.tab_widget.pack2(child, True, False)
        self.scroll_widget = builder.get_object("scroll_plot")
        self.motor_widget = MotorCreationWidget()
        builder.connect_signals(self)


    def configure_plot(self, matplotlib_plot):
        canvas = FigureCanvas(matplotlib_plot)
        canvas.set_size_request(800, 600)

        vbox = Gtk.VBox()
        toolbar = NavigationToolbar(canvas, self.scroll_widget)
        vbox.pack_start(toolbar, False, False, 0)
        vbox.pack_start(canvas, True, True, 0)

        self.button = Gtk.Button(label="Update plot")
        label = self.button.get_child()
        label.set_markup("<span size='x-large' foreground='red'><b>Update plot</b></span>")
        self.button.connect("clicked", self.user_asked_for_update)
        self.button.set_halign(Gtk.Align.CENTER)
        self.button.set_valign(Gtk.Align.START)
        self.button.set_margin_top(50)

        overlay = Gtk.Overlay()
        self.scroll_widget.add(overlay)
        overlay.add(vbox)
        overlay.add_overlay(self.button)

        self.button.set_visible(False)

    def plot_need_update(self):
        self.button.set_visible(True)

    def user_asked_for_update(self, *args):
        self.button.set_visible(False)
        self.update_plot()

    def update_library(self, library_data : "dict"):
        ''' Called when the main window library parameter has been updated '''
        self.motor_widget.update_library(library_data)

    def update_plot(self):
        pass



import typing as tp
import numpy as np
from matplotlib.figure import Figure
import matplotlib as mpl
import matplotlib.cm as mcolormaps

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib, GdkPixbuf

from .abstract_tab import AbstractTab
from .utils import *
from ..physics.battery import get_battery_state
from ..physics.motor import Motor

from ..ressources import get_ressource_path

DISPLAYED_CTES = [('Km, articular', 'Nm/\u221AW', lambda m: f"{m.K_m_art:.3f}"),
                  ('', 'W /Nm\u00B2', lambda m: f"{1 / m.K_m_art**2:.3f}"),
                  ('Power (non defluxing)', 'W', lambda m: f"{m.w_max_at_max_torque * m.iq_max * m.kt_q_art:.1f}"),
                  ('Ke', 'V/rad.s', lambda m: f"{m.ke:.3f}"),
                  ('Ktq, articular', 'Nm/A', lambda m: f"{m.kt_q_art:.3f}"),
                  ('Max torque, articular', 'Nm', lambda m: f"{m.tau_max:.1f}"),
                  ('No load speed, articular', 'rad/s', lambda m: f"{m.w_max_no_load:.1f}"),
                  ('', 'rpm', lambda m: f"{30 / np.pi * m.w_max_no_load:.0f}"),
                  ('Max speed @ max torque', 'rad/s', lambda m: f"{m.w_max_at_max_torque:.1f}"),
                  ('', 'rpm', lambda m: f"{30 / np.pi * m.w_max_at_max_torque:.0f}"),
                  ('R', 'Ohm', lambda m: f"{m.R:.3f}")]

class SingleMotorPerfTab(AbstractTab):
    '''
    GUI tab: look at the caracteristics of a single motor
    '''
    def __init__(self):
        builder = Gtk.Builder()
        builder.add_from_file(get_ressource_path("single_motor_tab.glade"))
        super().__init__("Single motor analysis", builder.get_object("side_box"))

        # Get the widgets we need from the builder.
        self.spin_bat_res = builder.get_object("spin_bat_res")
        self.spin_nominal_T = builder.get_object("spin_nominal_T")
        self.spin_flux_var = builder.get_object("spin_flux_var")
        self.spin_R_var = builder.get_object("spin_R_var")
        self.spin_rotor = builder.get_object("spin_rotor_temp")
        self.spin_stator = builder.get_object("spin_stator_temp")

        self.motors = [Motor(1, 0.01, 0.0001, 1.0, 1.0, 1.0, 1.0, 1.0) for _ in range(2)]

        self.label_specs = []
        for p in ["nominal_", "thermal_"]:
            self.label_specs.append(builder.get_object(f"{p}spec"))

        spec_legend = builder.get_object("spec_legend")
        spec_unit = builder.get_object("spec_unit")
        spec_legend.set_text("\n".join([d[0] for d in DISPLAYED_CTES]))
        spec_unit.set_text("\n".join([d[1] for d in DISPLAYED_CTES]))

        box = builder.get_object("side_box")
        box.pack_start(self.motor_widget.top_frame, False, False, 0)
        box.reorder_child(self.motor_widget.top_frame, 0)
        self.motor_widget.connect("motor_updated", self.input_updated)

        self.mpl_fig = Figure(figsize=(5, 4), dpi=100)
        ax = self.mpl_fig.add_subplot()
        self.mpl_fig.subplots_adjust(left=0.08, bottom=0.15, right=1.0, top=0.97, wspace=0, hspace=0)
        self.cbar = self.mpl_fig.colorbar(mpl.cm.ScalarMappable(mpl.colors.Normalize(vmin=0, vmax=1000)))

        self.configure_plot(self.mpl_fig)

        combo_box = builder.get_object("plot_type")
        self.change_plot_type(combo_box)
        builder.connect_signals(self)

    def input_updated(self, *args, **kargs):
        self.battery_resistance = self.spin_bat_res.get_value()

        self.motors[0] = self.motor_widget.motor
        To = self.spin_nominal_T.get_value()
        fvar = self.spin_flux_var.get_value() / 100.0
        Rvar = self.spin_R_var.get_value() / 100.0
        Tr = self.spin_rotor.get_value()
        Ts = self.spin_stator.get_value()
        self.motors[1].update_constants(
            n = 2 * self.motors[0].np,
            R = self.motors[0].R * (1 + Rvar * (Ts - To)),
            L = self.motors[0].L,
            ke = self.motors[0].ke * (1 - fvar * (Tr - To)),
            iq_max = self.motors[0].iq_max,
            iq_nominal = self.motors[0].iq_nominal,
            U = self.motors[0].U,
            reduction_ratio = self.motors[0].rho)

        for i in range(2):
            text = "\n".join([d[2](self.motors[i]) for d in DISPLAYED_CTES])
            self.label_specs[i].set_text(text)

        self.plot_need_update()

    def update_plot(self):
        # Use this for scaling & legend
        plot_caracteristics(self.mpl_fig.gca(), [self.motors[1]], ["C2"])
        # Next replot the curves in the right order
        plot_motor_caracteristic(self.mpl_fig.gca(), self.motors[0], "k")
        plot_motor_caracteristic(self.mpl_fig.gca(), self.motors[1], "C2")

        mot = self.motors[1]

        max_plot_speed = min(2 * mot.w_max_no_load, mot.compute_max_speed_deflux(0.0))
        w = np.linspace(0, max_plot_speed, 200)

        tau = np.linspace(0, mot.tau_max, 200)
        w_grid, tau_grid = np.meshgrid(w, tau)

        plot_surface = np.full(w_grid.shape, -np.inf)

        # Select plot content
        if self.plot_type[0] == "meca":
            plot_func = lambda t, w: t * w
        elif self.plot_type[0] == "thermal":
            plot_func = lambda t, w: mot.compute_thermal_power(t, w)
        elif self.plot_type[0] == "power":
            plot_func = lambda t, w: t * w + mot.compute_thermal_power(t, w)
        elif self.plot_type[0] == "efficiency":
            plot_func = lambda t, w: t * w / (t * w + mot.compute_thermal_power(t, w)) * 100
        elif self.plot_type[0] == "battery":
            plot_func = lambda t, w: get_battery_state(mot.U, self.battery_resistance,t * w + mot.compute_thermal_power(t, w))[0]

        for i in range(len(w_grid)):
            mask = w_grid[i] <= mot.compute_max_speed_deflux(tau_grid[i])
            plot_surface[i][mask] = plot_func(tau_grid[i][mask], w_grid[i][mask])
        ax = self.mpl_fig.gca()
        cm = mcolormaps.get_cmap("RdBu")
        cm = cm.reversed()
        cm.set_over('w')
        cm.set_under('#A0A0A0')
        ax.grid(False)
        q = ax.pcolormesh(w_grid, tau_grid, plot_surface, cmap=cm, shading='gouraud', rasterized=True)
        # Hide cursor data, the interpolation makes it wrong anyway.
        q.get_cursor_data = lambda event: None

        def format_coord(w, t):
            tail = ""
            if w < mot.compute_max_speed_deflux(t):
                tail = f", {self.plot_type[1]}: {plot_func(t, w):.1f}{self.plot_type[2]}"
            return f"Velocity: {w:.1f}rad/s ({w * 30 / np.pi:.1f}rpm), Torque: {t:.1f}Nm" + tail
        ax.format_coord = format_coord

        ax = self.cbar.ax
        ax.clear()
        self.cbar = self.mpl_fig.colorbar(mpl.cm.ScalarMappable(mpl.colors.Normalize(vmin=np.nanmin(plot_surface[plot_surface != -np.inf]), vmax=np.nanmax(plot_surface)), cmap=cm), cax = ax)
        ax.set_ylabel(self.plot_type[1])
        if self.plot_type[0] == "battery":
            sa = ax.secondary_yaxis(0.0,
                                    functions=(lambda u : (mot.U - u) / self.battery_resistance, lambda i : mot.U - self.battery_resistance * i))
            sa.set_ylabel('Battery current (A)', fontsize=10)

        self.mpl_fig.canvas.draw()

    def change_plot_type(self, combo_box):
        """
        Change the type of plot asked for
        """
        model = combo_box.get_model()
        self.plot_type = model[combo_box.get_active_iter()][1:]
        self.spin_bat_res.set_sensitive(self.plot_type[0] == "battery")
        self.plot_need_update()